using System;
using System.Collections.Generic;
using System.Text;
using ImageSharp.Images;
using ImageSharp.Maths;
using ImageSharp.Filtering.Linear;
using ImageSharp.HistOps;

namespace ImageSharp.Segmentations.Binarizations
{
    class Thresholding : IGrayImgApplicable
    {
        private byte threshold = 128;

        public Thresholding() { }
        public Thresholding(byte threshold)
        {
            this.threshold = threshold;
        }

        public void Apply(ImageGray img)
        {
            for (int y = 0; y < img.Height; y++)
            {
                for (int x = 0; x < img.Width; x++)
                {
                    // Wenn Pixel Schwellwert bersteigt, dann setzen, ansonsten entfernen
                    img.Data[y, x] = (byte)((img.Data[y, x] > this.threshold) ? 255 : 0);
                }
            }
        }
    }

    class OtsuThresholding : IGrayImgApplicable
    {
        public void Apply(ImageGray img)
        {
            img.Apply(new Thresholding((byte)GetOtsuThreshold(img)));
        }

        #region altes Verfahren
        //public int GetOtsuThreshold(ImageGray img)
        //{
        //    // calculate histogram
        //    int[] hist = HistogramOps.CalcHist(img);
            
        //    // calculate probability of the single gray values
        //    float[] p = gvalProb(img, hist);
            
        //    // calculate arithmetic gray value average
        //    float avg = gvalAvg(p);

        //    // find Otsu threshold
        //    float[] q = new float[256];
        //    int imax = -1;
        //    for (int t = 0; t <= 255; t++)
        //    {
        //        // calculate probabilities for pixels in both classes
        //        float p_k0 = pk0(p, t);
        //        float p_k1 = pk1(p, t);

        //        if (p_k0 > 0 && p_k1 > 0)
        //        {
        //            // calculate average gray value in both classes
        //            float avg_k0 = avgk0(p, p_k0, t);
        //            float avg_k1 = avgk1(p, p_k1, t);

        //            // calculate variance in the classes
        //            float var_k0 = vark0(p, avg_k0, p_k0, t);
        //            float var_k1 = vark1(p, avg_k1, p_k1, t);

        //            // calculate variance between the classes
        //            float varBetw = p_k0 * (avg_k0 - avg) * (avg_k0 - avg) + p_k1 * (avg_k1 - avg) * (avg_k1 - avg);
        //            float varIn = p_k0 * var_k0 + p_k1 * var_k1;

        //            // get quotient; if it's bigger than the highest quotient -> better threshold found with current t
        //            q[t] = varBetw / varIn;
        //            //q[t] = varBetw;
        //            if (imax == -1 || q[t] > q[imax])
        //                imax = t;
        //        }
        //    }

        //    return imax;
        //}

        //// Returns the probability for pixels in class 0.
        //private float pk0(float[] p, int t)
        //{
        //    float val = 0.0f;
        //    for (int g = 0; g <= t; g++)
        //    {
        //        val += p[g];
        //    }
        //    return val;
        //}

        //// Returns the probability for pixels in class 1.
        //private float pk1(float[] p, int t)
        //{
        //    float val = 0.0f;
        //    for (int g = t + 1; g < p.Length; g++)
        //    {
        //        val += p[g];
        //    }
        //    return val;
        //}

        //// Get average gray value of class 0.
        //private float avgk0(float[] p, float p_k0, int t)
        //{
        //    float avg_k0 = 0.0f;
        //    for (int i = 0; i <= t; i++)
        //    {
        //        avg_k0 += p[i] * (i + 1) / p_k0;
        //    }
        //    avg_k0 -= 1.0f;

        //    return avg_k0;
        //}

        //// Get average gray value of class 1.
        //private float avgk1(float[] p, float p_k1, int t)
        //{
        //    float avg_k1 = 0.0f;
        //    for (int i = t + 1; i <= 255; i++)
        //    {
        //        avg_k1 += p[i] * (i + 1) / p_k1;
        //    }
        //    avg_k1 -= 1.0f;

        //    return avg_k1;
        //}

        //// Returns variance of pixels in class 0.
        //private float vark0(float[] p, float avg_k0, float p_k0, int t)
        //{
        //    float val = 0.0f;
        //    for (int g = 0; g <= t; g++)
        //    {
        //        val += (g - avg_k0) * (g - avg_k0) * p[g] / p_k0;
        //    }
        //    return val;
        //}

        //// Returns variance of pixels in class 1.
        //private float vark1(float[] p, float avg_k1, float p_k1, int t)
        //{
        //    float val = 0.0f;
        //    for (int g = t + 1; g < p.Length; g++)
        //    {
        //        val += (g - avg_k1) * (g - avg_k1) * p[g] / p_k1;
        //    }
        //    return val;
        //}
        #endregion

        public int GetOtsuThreshold(ImageGray img)
        {
            // calculate probability of the single gray values
            float[] p = gvalProb(img);

            // calculate arithmetic gray value average
            float avg = gvalAvg(p);

            // find Otsu threshold
            float[] p_k0 = new float[256];
            float[] avg_k0 = new float[256];
            float[] avg_k1 = new float[256];
            double[] q = new double[256];
            int imax = -1;
            p_k0[0] = p[0];
            for (int t = 1; t <= 255; t++)
            {
                p_k0[t] = p_k0[t - 1] + p[t];
                if (p_k0[t] > 0 && p_k0[t] < 1)
                {
                    avg_k0[t] = (p_k0[t - 1] * avg_k0[t - 1] + t * p[t]) / p_k0[t];
                    avg_k1[t] = (avg - p_k0[t] * avg_k0[t]) / (1f - p_k0[t]);

                    // if between class variance is bigger than the highest variance -> better threshold found with current t
                    q[t] = p_k0[t] * (1f - p_k0[t]) * Math.Pow(avg_k0[t] - avg_k1[t], 2.0);
                    if (imax == -1 || q[t] > q[imax])
                        imax = t;
                }
            }

            return imax;
        }

        // Get probabilities of the gray values.
        private float[] gvalProb(ImageGray img)
        {
            int[] hist = HistogramOps.CalcHist(img);

            float[] p = new float[256];
            float pxs = img.Height * img.Width;
            for (int i = 0; i < hist.Length; i++)
            {
                p[i] = hist[i] / pxs;
            }

            return p;
        }

        // Get gray value average.
        private float gvalAvg(float[] p)
        {
            float avg = 0.0f;
            for (int i = 0; i <= 255; i++)
            {
                avg += p[i] * (i + 1f);
            }
            avg -= 1.0f;

            return avg;
        }
    }

    /// <summary>
    /// Extraktion von Nulldurchlufen im bergebenen Rechteck-Bereich, wobei der Wert 0 (Nulldurchlaufspunkt)
    /// dem bergebenen Grauwert entspricht.
    /// </summary>
    class ZeroCrossings : IGrayImgApplicable
    {
        private byte threshold = 128;

        public ZeroCrossings() { }
        /// <summary>
        /// Konstruktor.
        /// </summary>
        /// <param name="threshold">Grauwert, der als Nullwert zur Separierung angesehen werden soll</param>
        public ZeroCrossings(byte threshold)
        {
            this.threshold = threshold;
        }

        public void Apply(ImageGray img)
        {
            for (int y = 0; y < img.Height; y++)
            {
                for (int x = 0; x < img.Width; x++)
                {
                    int x1 = (x == 0) ? x + 1 : x - 1;
                    int y1 = (y == img.Height - 1) ? y - 1 : y + 1;
                    
                    // Vorzeichen des aktuellen Pixels erhalten
                    int sign = MathCalc.SignOwn(img.Data[y, x] - this.threshold);
                    // Vorzeichen des rechten bzw. linken Pixels
                    int signX = MathCalc.SignOwn(img.Data[y, x1] - this.threshold);
                    // Vorzeichen des unteren bzw. oberen Pixels
                    int signY = MathCalc.SignOwn(img.Data[y1, x] - this.threshold);
                    byte val = 255;

                    // Wenn kein Nulldurchlauf (alle Vorzeichen gleich), dann 0 setzen
                    if (sign == signX && sign == signY)
                    {
                        val = 0;
                    }

                    // Wert in Daten-Array eintragen
                    img.Data[y, x] = val;
                }
            }
        }
    }

    class ADBinarization : IGrayImgApplicable
    {
        private enum Methods
        {
            ON_Center,
            OFF_Center
        }

        private class MaskInfo
        {
            public int Sum = -1;
            public int SumUpperRow = 0;
            public int ColSpan = 0;
            public int PxCount = 0;
        };

        private class GanglionInfo
        {
            public float S = 0.0f;  // Surround-Reaktion einer Ganglionzelle
            public float C = 0.0f;  // Center-Reaktion einer Ganglionzelle
        };

        public class GanglionCell
        {
            public int Sr = 0;  // Surround-Radius einer Ganglionzelle
            public int Cr = 0;  // Center-Radius einer Ganglionzelle
            public GanglionCell() { }
            public GanglionCell(int sr, int cr)
            {
                this.Sr = sr;
                this.Cr = cr;
            }
        };

        private GanglionCell[] gcells = null;

        public ADBinarization()
        {
            this.gcells = new GanglionCell[4];
            this.gcells[0] = new GanglionCell(7, 1);
            this.gcells[1] = new GanglionCell(3, 1);
            this.gcells[2] = new GanglionCell(3, 0);
            this.gcells[3] = new GanglionCell(1, 0);
        }

        public ADBinarization(GanglionCell[] gcells)
        {
            this.gcells = gcells;
        }

        public void Apply(ImageGray img)
        {
            // HVS-Schwellwertverfahren anwenden
            ImageGray img1 = thresholdHVS(img);

            // Bilddaten bertragen
            img.Data = img1.Data;
        }

        private ImageGray thresholdHVS(ImageGray img)
        {
            // erhalte zunchst die Aktivierungen der Ganglienzellen
            ImageGray img1 = getGangliaActivations(img, Methods.OFF_Center);

            // berechne auf Aktivierungsbild Schwellwert nach Otsu und binarisiere
            img1.Apply(new OtsuThresholding());

            return img1;
        }

        private ImageGray getGangliaActivations(ImageGray img, Methods method)
        {
            ImageGray nimg = new ImageGray(img.Width, img.Height);

            int i, j, k;
            // der folgende Algorithmus sortiert doppelte Masken aus cellRadius aus, sortiert
            // das Array dieser Masken und berechnet dann die lokalen Summen jeder Maske
            // 1.) Array fr alle zu erstellenden Masken anlegen und sortieren
            int[] masks1 = new int[2 * gcells.Length];
            for (i = 0; i < gcells.Length; i++)
            {
                masks1[2 * i] = gcells[i].Cr;
                masks1[2 * i + 1] = gcells[i].Sr;
            }
            Array.Sort(masks1);
            // 2.) Duplikate ausschlieen
            for (i = 0, j = 1; j < masks1.Length; j++)
            {
                if (masks1[i] != masks1[j])
                {
                    i++;
                    masks1[i] = masks1[j];
                }
            }
            int mcount = i + 1;
            int[] masks = new int[mcount];
            for (j = 0; j < mcount; j++)
            {
                masks[j] = masks1[j];
            }

            // 3.) entsprechend der Masken Arrays von MaskInfo-Objekten erstellen;
            // diese beinhalten neben der Maskensumme fr ein Pixel Zusatzinformationen,
            // die zur effizienten Berechnung bentigt werden
            MaskInfo[,,] carrier = new MaskInfo[mcount, img.Height, img.Width];
            for (i = 0; i < mcount; i++)
                for (j = 0; j < img.Height; j++)
                    for (k = 0; k < img.Width; k++)
                        carrier[i, j, k] = new MaskInfo();

            // 4.) zu jeder Maske die Summe fr jedes Pixel holen
            getCarrierInformation(img, masks, carrier);

            // nun fr jede Ganglienzelle zunchst herausfinden, welche Masken ihrem
            // Center und Surround entsprechen;
            // dann die damit die Aktivierungen der Ganglienzelle fr jedes Pixel berechnen
            GanglionInfo[, ,] ginfos = new GanglionInfo[gcells.Length, img.Height, img.Width];
            for (i = 0; i < gcells.Length; i++)
            {
                // 1.) berechne Maskenindex von Center und Surround
                int inC = 0;
                int inS = 0;
                for (j = 0; inS == 0 && j < mcount; j++)
                {
                    if (masks[j] == gcells[i].Cr)
                    {
                        inC = j;
                    }
                    else if (masks[j] == gcells[i].Sr)
                    {
                        inS = j;
                    }
                }

                // 2.) erstelle GanglionInfo-Objekte, welche 
                // die Center- und Surround-Summen speichern
                for (j = 0; j < img.Height; j++)
                {
                    for (k = 0; k < img.Width; k++)
                    {
                        ginfos[i, j, k] = new GanglionInfo();
                        int c = carrier[inC, j, k].Sum;
                        int pxc = carrier[inC, j, k].PxCount;
                        int s = carrier[inS, j, k].Sum;
                        int pxs = carrier[inS, j, k].PxCount;
                        ginfos[i, j, k].C = c / (float)pxc;
                        ginfos[i, j, k].S = (s - c) / (float)(pxs - pxc);
                    }
                }
            }

            // nun die Aktivierungen der Ganglienzellen berechnen
            for (int y = 0; y < img.Height; y++)
            {
                for (int x = 0; x < img.Width; x++)
                {
                    // erhalte zunchst die Informationen der hchsten Aktivierung
                    int index = getMaxContrastGanglionInfo(img, x, y, ginfos, method);

                    // berechne nun die Aktivierung
                    int act = calcGanglionActivation(ginfos[index, y, x], method);
                    nimg.Data[y, x] = (byte)(255-act);
                }
            }

            return nimg;
        }

        private int getMaxContrastGanglionInfo(ImageGray img, int x, int y, GanglionInfo[, ,] ginfos, Methods method)
        {
            // Objekte fr die bisher maximale Aktivierung
            int imax = 0;
            float gimaxcon = (method == Methods.ON_Center) ? (ginfos[0, y, x].C - ginfos[0, y, x].S) : (ginfos[0, y, x].S - ginfos[0, y, x].C);
            for (int i = 1; i < ginfos.GetLength(0); i++)
            {
                GanglionInfo gi = ginfos[i, y, x];
                float gicon = (method == Methods.ON_Center) ? (gi.C - gi.S) : (gi.S - gi.C);
                if (gicon > gimaxcon)
                {
                    imax = i;
                    gimaxcon = gicon;
                }
            }

            // Index der Ganglienzellen mit dem hchsten Kontrast zurckgeben
            return imax;
        }

        private int calcGanglionActivation(GanglionInfo ginfo, Methods method)
        {
            int offset = 10; // angegebener "optimaler" Offset
            int act = 0;

            float fmax = (ginfo.S - ginfo.C);

            if (method == Methods.ON_Center)
                fmax = -fmax;

            // berechne Aktivierung nach vorgegebener Formel
            if (fmax > 0)
            {
                float A = (method == Methods.ON_Center ? ginfo.C : ginfo.S) + offset;
                act = Math.Min(255, (int)((fmax * (A + 255)) / (A + fmax)));
            }

            return act;
        }

        private void getCarrierInformation(ImageGray img, int[] masks, MaskInfo[,,] carrier)
        {
            for (int k = 0; k < masks.Length; k++)
            {
                int i, j;
                // wenn der Radius 0 ist, dann ist es besser einfach nur die
                // Werte zu mappen statt kompliziert rechnen zu wollen
                if (masks[k] == 0)
                {
                    for (i = 0; i < img.Height; i++)
                    {
                        for (j = 0; j < img.Width; j++)
                        {
                            carrier[k, i, j].Sum = img.Data[i, j];
                            carrier[k,i,j].PxCount = 1;
                        }
                    }
                    continue;
                }

                // Bild muss schon eine gewisse Dimensionierung haben
                if (img.Width <= 2 * masks[k] || img.Height <= 2 * masks[k])
                {
                    continue;
                }

                // berlappende Summen in Breite der aktuellen Maske
                // fr jede Zeile berechnen
                for (i = 0; i < img.Height; i++)
                {
                    int toSub = img.Data[i, 0];
                    int curSum = 0;
                    for (j = 0; j < img.Width + masks[k]; j++)
                    {
                        // wenn rechts noch ein Pixel kommt, dann zur aktuellen Summe addieren
                        if (j < img.Width)
                        {
                            curSum += img.Data[i, j];
                        }
                        if (j >= masks[k])
                        {
                            // wenn nicht mehr am linken Rand, dann Pixel links neben aktueller
                            // Maskenposition subtrahieren und neues linkes Pixel berechnen
                            if (j > 2 * masks[k])
                            {
                                curSum -= toSub;
                                toSub = img.Data[i, j - 2 * masks[k]];
                            }
                            // Zeilensumme fr das Pixel speichern
                            carrier[k, i, j - masks[k]].SumUpperRow = curSum;
                            // fr die erste Zeile die x-Ausdehnung der Maske speichern,
                            // => Anzahl involvierter Pixel fr die jeweilige Position
                            if (i == 0)
                            {
                                carrier[k, 0, j - masks[k]].ColSpan = (j >= 2 * masks[k]) ? ((j < img.Width) ? (2 * masks[k] + 1) : (2 * masks[k] - j + img.Width)) : (j + 1);
                            }
                        }
                    }
                }

                // jetzt mit den berechneten Zeilensummen die eigentliche Maskensumme berechnen
                // -> prinzipiell selbes Verfahren wie oben
                for (j = 0; j < img.Width; j++)
                {
                    int toSub = carrier[k, 0, j].SumUpperRow;
                    int curSum = 0;
                    for (i = 0; i < img.Height + masks[k]; i++)
                    {
                        if (i < img.Height)
                        {
                            curSum += carrier[k, i, j].SumUpperRow;
                        }
                        if (i >= masks[k])
                        {
                            if (i > 2 * masks[k])
                            {
                                curSum -= toSub;
                                toSub = carrier[k, i - 2 * masks[k], j].SumUpperRow;
                            }
                            carrier[k, i - masks[k], j].Sum = curSum;
                            carrier[k, i - masks[k], j].PxCount = carrier[k, 0, j].ColSpan * ((i >= 2 * masks[k]) ? ((i < img.Height) ? (2 * masks[k] + 1) : (2 * masks[k] - i + img.Height)) : (i + 1));
                        }
                    }
                }
            }
        }
    }
}
