using System;
using System.Collections.Generic;
using System.Text;
using ImageSharp.Images;
using ImageSharp.HistOps;
using System.Drawing;

namespace ImageSharp.GrayValueTransforms
{
    /// <summary>
    /// Fhrt eine Gamma-Korrektur mit den angegebenen Parametern durch.
    /// </summary>
    class GammaCorrection : IGrayImgApplicable
    {
        private float gamma = 0;
        private int a0 = 0;

        /// <summary>
        /// Konstruktor.
        /// </summary>
        /// <param name="gamma">Gamma-Wert, der fr die Korrektur verantwortlich ist</param>
        /// <param name="a0">Wert, bis zu dem der Wert linear abgebildet werden soll</param>
        public GammaCorrection(float gamma, int a0)
        {
            this.gamma = gamma;
            this.a0 = a0;
        }

        public void Apply(ImageGray img)
        {
            // temporre Werte berechnen
            double a1 = this.a0 / 255.0f;
            double s = this.gamma / (a1 * (this.gamma - 1.0) + Math.Pow(a1, 1.0 - this.gamma));
            double d = 1.0f / (Math.Pow(a1, this.gamma) * (this.gamma - 1.0) + 1.0) - 1.0;

            // LUT allokieren
            byte[] lut = new byte[256];

            // LUT erzeugen
            for (int i = 0; i <= 255; i++)
            {
                double lutVal = 0.0;
                float a = i / 255.0f;
                // unteren Bereich linear abbilden, oberen aus Gamma-Korrektur
                if (a <= a1)
                    lutVal = s * a;
                else
                    lutVal = (1.0 + d) * Math.Pow(a, this.gamma) - d;
                lut[i] = (byte)Math.Round(lutVal * 255.0);
            }

            // LUT anwenden
            img.Apply(new LUTApplication(lut));
        }
    }

    /// <summary>
    /// Kontrast-Erhhung des Bildes.
    /// </summary>
    class ContrastStretching : IGrayImgApplicable
    {
        private Point p1;
        private Point p2;

        public ContrastStretching(byte grayOld1, byte grayNew1, byte grayOld2, byte grayNew2)
        {
            // Zwischenpunkte fr die Transformation definieren
            this.p1 = new Point(grayOld1, grayNew1);
            this.p2 = new Point(grayOld2, grayNew2);
            if (grayOld2 < grayOld1)
            {
                Point temp = this.p1;
                this.p1 = this.p2;
                this.p2 = temp;
            }
        }

        public void Apply(ImageGray img)
        {
            (new FunctionApplication(contrStretchFunc)).Apply(img);
        }

        // Transformationsfunktion zur Kontrasterhhung.
        // Transformiert den bergebenen Grauwert in den neuen Wert.
        private byte contrStretchFunc(byte pxValue)
        {
            // lineare Interpolation der Grauwerte auf die Linien der Punkte p1, p2 und der 
            // Punkte (0,0) und (255,255)
            if (pxValue < p1.X)
            {
                return (byte)(p1.Y / (double)p1.X * pxValue);
            }
            else if (pxValue > p2.X)
            {
                double m = (255.0 - p2.Y) / (255.0 - p2.X);
                return (byte)(m * pxValue + 255 - m * 255);
            }
            else
            {
                double m = (p2.Y - p1.Y) / (double)(p2.X - p1.X);
                return (byte)(m * pxValue + p2.Y - m * p2.X);
            }
        }
    }

    /// <summary>
    /// Kontrastanpassung mit den bergebenen Prozentwerten fr untere und obere Grenze.
    /// </summary>
    class ContrastAdaption : IGrayImgApplicable
    {
        private float pxPercLow = 0f;
        private float pxPercHigh = 0f;

        /// <summary>
        /// Kontrastanpassung mit den bergebenen Prozentwerten fr untere und obere Grenze.
        /// </summary>
        /// <param name="pxPercLow">Prozentwert fr untere Grenze</param>
        /// <param name="pxPercHigh">Prozentwert fr obere Grenze</param>
        public ContrastAdaption(float pxPercLow, float pxPercHigh)
        {
            this.pxPercLow = pxPercLow;
            this.pxPercHigh = pxPercHigh;
        }

        public void Apply(ImageGray img)
        {
            // Histogramm erzeugen
            int[] hist = HistogramOps.CalcHist(img);
            int threshLow = 0;
            int threshHigh = 255;

            // Pixelanzahl im Bildbereich
            int pxs = img.Width * img.Height;

            // oberen Schwellwertindex des Histogramms berechnen - der Bereich ber diesem muss
            // dem bergebenen oberen Prozentsatz der Pixel im Bildbereich entsprechen
            for (int i = 255, pxCount = 0; i >= 0; i--)
            {
                pxCount += hist[i];
                if (pxCount >= pxs * this.pxPercHigh / 100.0f)
                {
                    threshHigh = i;
                    break;
                }
            }

            // unteren Schwellwertindex des Histogramms berechnen - der Bereich unter diesem muss
            // dem bergebenen unteren Prozentsatz der Pixel im Bildbereich entsprechen
            for (int i = 0, pxCount = 0; i <= 255; i++)
            {
                pxCount += hist[i];
                if (pxCount >= pxs * this.pxPercLow / 100.0f)
                {
                    threshLow = i;
                    break;
                }
            }

            // Lookup-Tabelle durch lineare Interpolation des Bereiches zwischen unterem und oberem
            // Histogramm-Index erzeugen
            byte[] lut = new byte[256];
            for (int i = 0; i <= 255; i++)
            {
                if (i <= threshLow)
                    lut[i] = 0;
                else if (i >= threshHigh)
                    lut[i] = 255;
                else
                    lut[i] = (byte)((i - threshLow) * 255.0f / (threshHigh - threshLow));
            }

            // LUT anwenden
            img.Apply(new LUTApplication(lut));
        }
    }

    /// <summary>
    /// Anwendung der bergebenen Lookup-Tabelle auf ein Bild.
    /// </summary>
    class LUTApplication : IGrayImgApplicable
    {
        private byte[] lut = null;

        /// <summary>
        /// Konstruktor.
        /// </summary>
        /// <param name="lut">Anzuwendende LUT</param>
        public LUTApplication(byte[] lut)
        {
            this.lut = lut;
        }

        public void Apply(ImageGray img)
        {
            for (int y = 0; y < img.Height; y++)
            {
                for (int x = 0; x < img.Width; x++)
                {
                    // Alten Grauwert als Index der LUT nutzen und deren Wert auslesen
                    img.Data[y, x] = this.lut[img.Data[y, x]];
                }
            }
        }
    }

    /// <summary>
    /// Wendet die bergebene Grauwert-Transformations-Funktion auf das Bild im angegebenen Bildbereich an.
    /// </summary>
    class FunctionApplication : IGrayImgApplicable
    {
        /// <summary>
        /// Delegate fr eine Grauwert-Transformations-Funktion. Diese bernimmt einen Grauwert und
        /// gibt dessen transformierten Wert zurck.
        /// </summary>
        /// <param name="pxValue">Zu transformierender Grauwert</param>
        /// <returns>Tansformierter Grauwert</returns>
        public delegate byte TransformFunction(byte pxValue);

        private TransformFunction transFunc = null;

        /// <summary>
        /// Konstruktor.
        /// </summary>
        /// <param name="method">Grauwert-Transformations-Funktion</param>
        public FunctionApplication(TransformFunction transFunc)
        {
            this.transFunc = transFunc;
        }

        public void Apply(ImageGray img)
        {
            if (this.transFunc == null)
                return;
            for (int y = 0; y < img.Height; y++)
            {
                for (int x = 0; x < img.Width; x++)
                {
                    img.Data[y, x] = this.transFunc(img.Data[y, x]);
                }
            }
        }
    }

    class GrayValueInversion : IGrayImgApplicable
    {
        public void Apply(ImageGray img)
        {
            for (int y = 0; y < img.Height; y++)
                for (int x = 0; x < img.Width; x++)
                    img.Data[y, x] = (byte)(255 - img.Data[y, x]);
        }
    }
}
