using System;
using System.Collections.Generic;
using System.Text;
using ImageSharp.Images;
using ImageSharp.Maths;
using ImageSharp.Filtering.Linear;

namespace ImageSharp.Filtering.NonLinear
{
    interface NonLinearFilter : IFilter
    {
        // void
    }

    class MedianFilter : NonLinearFilter
    {
        public void Apply(ImageGray img)
        {
            throw new Exception("The method or operation is not implemented.");
        }
    }

    /// <summary>
    /// Unscharf-Maskierung des Grauwertbildes mit der angegebenen Standardabweichung fr die
    /// Gauss-Funktion sowie der Intensitt der Maskierung im bergebenen Bildbereich.
    /// </summary>
    class UnsharpMasking : NonLinearFilter
    {
        private double sigma = 0;
        private double intensity = 0;

        /// <summary>
        /// Konstruktor.
        /// </summary>
        /// <param name="sigma">Standardabweichung der Gauss-Funktion</param>
        /// <param name="intensity">Intensitt des Filters</param>
        public UnsharpMasking(double sigma, double intensity)
        {
            this.sigma = sigma;
            this.intensity = intensity;
        }

        public void Apply(ImageGray img)
        {
            // erzeuge Kopie des Bildes und tiefpassgefiltertes Bild
            ImageGray lowpass = new ImageGray(img);
            lowpass.Apply(new LinearFilter(MathCalc.GaussMask(this.sigma)));

            // Subtrahiere das tiefpassgefilterte Bild vom Originalbild
            ImageGray sub = Arithmetics.Sub(img, lowpass);
            // Addiere das Ergebnis mit der angegebenen Intensitt fr das Ergebnisbild zum Original
            ImageGray imgGray = Arithmetics.Add(sub, img, 1.0, this.intensity);
            img.Data = imgGray.Data;
        }
    }

    /// <summary>
    /// Highboost-Filterung des Bildes mit der bergebenen Standardabweichung.
    /// </summary>
    class HighboostFilter : NonLinearFilter
    {
        private double sigma = 1.0;

        public HighboostFilter() { }
        /// <summary>
        /// Konstruktor.
        /// </summary>
        /// <param name="sigma">Standardabweichung fr die Gauss-Funktion</param>
        public HighboostFilter(double sigma)
        {
            this.sigma = sigma;
        }

        public void Apply(ImageGray img)
        {
            ImageGray lowpass = new ImageGray(img);
            lowpass.Apply(new LinearFilter(MathCalc.GaussMask(this.sigma)));
            ImageGray imgGray = Arithmetics.Sub(img, lowpass, 2.0, 1.0);
            img.Data = imgGray.Data;
        }
    }

    /// <summary>
    /// Highboost-Filterung unter Anwendung des linearen Gauss-Filters.
    /// </summary>
    class HighboostFilter2 : NonLinearFilter
    {
        public void Apply(ImageGray img)
        {
            // Tiefpassgefiltertes Bild erzeugen
            ImageGray lowpass = new ImageGray(img);
            lowpass.Apply(LinearFilters.Gaussian);
            // Das Bild mit einem Faktor von 2 belegen und tiefpassgefiltertes Bild subtrahieren
            ImageGray imgGray = Arithmetics.Sub(img, lowpass, 2.0, 1.0);
            img.Data = imgGray.Data;
        }
    }

    class ISEFFilter : NonLinearFilter
    {
        private float smoothingFactor = 0f;

        public ISEFFilter(float smoothingFactor)
        {
            this.smoothingFactor = smoothingFactor;
        }

        public void Apply(ImageGray img)
        {
            // ISEF-Bild berechnen
            float[,] isef = CalcISEFImg(img);

            // Ergebnis im Bild speichern
            for (int i = 0; i < img.Height; i++)
                for (int j = 0; j < img.Width; j++)
                    img.Data[i, j] = (byte)isef[i,j];
        }

        public float[,] CalcISEFImg(ImageGray img)
        {
            // Speicher fr Teilergebnis allokieren
            float[,] isef = new float[img.Height, img.Width];

            // Eingabebild mit ISEF gltten
            computeISEF(img.Data, isef);

            return isef;
        }

        // Rekursive Anwendung des ISEF-Filters
        // (Shen and Castan CVIGP Mrz 1992)
        private void computeISEF(byte[,] x, float[,] y)
        {
            float[,] A = new float[x.GetLength(0), x.GetLength(1)]; // kausale Komponente
            float[,] B = new float[x.GetLength(0), x.GetLength(1)]; // anti-kausale Komponente

            // wende Filter zuerst vertikal (auf die Zeilen) an
            applyISEFvertical(x, y, A, B);

            // wende Filter horizontal (auf die Spalten) an und verknpfe mit vertikalem Ergebnis
            applyISEFhorizontal(y, y, A, B);
        }

        // Vertikale (zeilenweise) Anwendung des ISEF-Filters
        private void applyISEFvertical(byte[,] x, float[,] y, float[,] A, float[,] B)
        {
            float b1, b2;
            int height = x.GetLength(0);
            int width = x.GetLength(1);

            b1 = (1.0f - this.smoothingFactor) / (1.0f + this.smoothingFactor);
            b2 = this.smoothingFactor * b1;

            // Randbedingungen berechnen
            for (int col = 0; col < width; col++)
            {
                A[0, col] = b1 * x[0, col];
                B[height - 1, col] = b2 * x[height - 1, col];
            }

            // kausale Komponente berechnen
            for (int row = 1; row < height; row++)
                for (int col = 0; col < width; col++)
                    A[row, col] = b1 * x[row, col] + this.smoothingFactor * A[row - 1, col];

            // anti-kausale Komponente berechnen
            for (int row = height - 2; row >= 0; row--)
                for (int col = 0; col < width; col++)
                    B[row, col] = b2 * x[row, col] + this.smoothingFactor * B[row + 1, col];

            // Randbedingung fr Ausgabe von erstem Filter
            for (int col = 0; col < width - 1; col++)
                y[height - 1, col] = A[height - 1, col];

            // nun berechne die Ausgabe des ersten Filters und speichere sie in y
            // ergibt sich aus der Summe von kausaler und anit-kausaler Komponente
            for (int row = 0; row < height - 1; row++)
                for (int col = 0; col < width; col++)
                    y[row, col] = A[row, col] + B[row + 1, col];
        }

        // Horizontale (spaltenweise) Anwendung des ISEF-Filters
        private void applyISEFhorizontal(float[,] x, float[,] y, float[,] A, float[,] B)
        {
            float b1, b2;
            int height = x.GetLength(0);
            int width = x.GetLength(1);

            b1 = (1.0f - this.smoothingFactor) / (1.0f + this.smoothingFactor);
            b2 = this.smoothingFactor * b1;

            // Randbedingungen berechnen
            for (int row = 0; row < height; row++)
            {
                A[row, 0] = b1 * x[row, 0];
                B[row, width - 1] = b2 * x[row, width - 1];
            }

            // berechne kausale Komponente
            for (int col = 1; col < width; col++)
                for (int row = 0; row < height; row++)
                    A[row, col] = b1 * x[row, col] + this.smoothingFactor * A[row, col - 1];

            // berechne anti-kausale Komponente
            for (int col = width - 2; col >= 0; col--)
                for (int row = 0; row < height; row++)
                    B[row, col] = b2 * x[row, col] + this.smoothingFactor * B[row, col + 1];

            // Randbedingung fr die Ausgabe des zweiten Filters
            for (int row = 0; row < height; row++)
                y[row, width - 1] = A[row, width - 1];

            // nun berechne die Ausgabe des zweiten Filters und speichere sie in y
            // ergibt sich aus der Summe von kausaler und anit-kausaler Komponente
            for (int row = 0; row < height; row++)
                for (int col = 0; col < width - 1; col++)
                    y[row, col] = A[row, col] + B[row, col + 1];
        }  
    }
}