using System;
using System.Collections.Generic;
using System.Text;
using ImageSharp.Images;
using ImageSharp.Filtering.NonLinear;

namespace ImageSharp.Detections.EdgeDetectors.Optimal
{
    /// <summary>
    /// Implementierung des optimalen Canny-Kantendetektors.
    /// </summary>
    class CannyDetector : IGrayImgApplicable
    {
        #region Private Attribute
        private float sigma = 1;                // Standardabweichung der Gauss-Funktion
        private byte thresholdLow = 0;          // anzuwendender unterer Schwellwert
        private byte thresholdHigh = 255;       // anzuwendender oberer Schwellwert
        private byte initThreshLow = 0;         // initial bergebener unterer Schwellwert
        private byte initThreshHigh = 255;      // initial bergebener oberer Schwellwert
    
        private const int MAX_MASK_SIZE = 20;   // max. Gre der Gauss-Maske
        private const int ORI_SCALE = 40;       // Skalierungsfaktor der Gradientenrichtung
        private const int MAG_SCALE = 20;       // Skalierungsfaktor der Gradientenstrke
        private float[,] gradOri;               // Feld zur Speicherung der Gradientenrichtungen
        private byte[,] gradMag;                // Feld zur Speicherung der Gradientenstrken
        private bool gradOnly = false;          // true, wenn nur Gradienteninformationen zu speichern sind
        #endregion

        #region ffentliche Eigenschaften
        /// <summary>
        /// Gibt an, ob die Gradientenstrke im Bild zu speichern ist anstatt dass im Anschluss
        /// noch ein Hysteresis-thresholding durchgefhrt wird. Auerdem wrde die Gradientenrichtung
        /// gespeichert werden. Wenn false, dann wird das Kantenbild binarisiert ber Hysteresis-thresholding.
        /// </summary>
        public bool GradOnly
        {
            get
            {
                return this.gradOnly;
            }
            set
            {
                this.gradOnly = value;
            }
        }

        /// <summary>
        /// Rckgabe des Arrays der Gradientenstrke.
        /// </summary>
        public byte[,] GradMagnitude
        {
            get
            {
                return this.gradMag;
            }
        }

        /// <summary>
        /// Rckgabe des Arrays der Gradientenrichtung - nur gesetzt, wenn GradOnly = true.
        /// </summary>
        public float[,] GradOrientation
        {
            get
            {
                return this.gradOri;
            }
        }
        #endregion

        #region ffentliche Methoden
        /// <summary>
        /// Erzeugt einen neuen Canny-Kantendetektor mit den bergebenen Parametern.
        /// </summary>
        /// <param name="sigma">Standardabweichung der Gauss-Funktion</param>
        /// <param name="threshLow">Unterer Schwellwert fr Hysteresis-Thresholding</param>
        /// <param name="threshHigh">Oberer Schwellwert fr Hysteresis-Thresholding</param>
        public CannyDetector(float sigma, byte threshLow, byte threshHigh)
        {
            this.sigma = sigma;
            this.initThreshLow = threshLow;
            this.initThreshHigh = threshHigh;
            this.thresholdLow = threshLow;
            this.thresholdHigh = threshHigh;
        }

        /// <summary>
        /// Erzeugt einen neuen Canny-Kantendetektor mit den bergebenen Parametern.
        /// </summary>
        /// <param name="sigma">Standardabweichung der Gauss-Funktion</param>
        /// <param name="threshLow">Unterer Schwellwert fr Hysteresis-Thresholding</param>
        /// <param name="threshHigh">Oberer Schwellwert fr Hysteresis-Thresholding</param>
        /// <param name="gradOnly">Gibt an, ob Gradienteninformationen auszugeben sind anstelle
        /// des binarisierten Kantenbildes</param>
        public CannyDetector(float sigma, byte threshLow, byte threshHigh, bool gradOnly)
            : this(sigma, threshLow, threshHigh)
        {
            this.gradOnly = gradOnly;
        }

        /// <summary>
        /// Wendet den Canny-Kantendetektor auf das bergebene Bild an.
        /// </summary>
        /// <param name="img">Bild, in dem Kanten detektiert werden sollen</param>
        public void Apply(ImageGray img)
        {
            // Teilbilder allokieren
            int width = 0;  // Breite der Gauss-Filtermaske
            float z = 0.0f;
            float[] gau = new float[MAX_MASK_SIZE], dgau = new float[MAX_MASK_SIZE];
            float[,] ix = new float[img.Height, img.Width];
            float[,] iy = new float[img.Height, img.Width];
            float[,] dx = new float[img.Height, img.Width];
            float[,] dy = new float[img.Height, img.Width];
            byte[,] mag = new byte[img.Height, img.Width];
            this.gradOri = new float[img.Height, img.Width];

            // Erzeuge Gauss-Filtermaske und Maske der 1. Ableitung von Gauss
            for (int i = 0; i < MAX_MASK_SIZE; i++)
            {
                gau[i] = meanGauss(this.sigma, i);
                if (gau[i] < 0.005f) // brich ab, wenn Fehler gering genug wird
                {
                    width = i;
                    break;
                }
                dgau[i] = dGauss(i, this.sigma);
            }
            
            // Bild mit Gauss falten -> jeweils in x- und y-Richtung
            convolve_separated(img, gau, width, ix, iy);
            
            // Gauss-Bilder mit Ableitungsmasken falten
            convolve_separated_dxy(ix, dgau, width, dx, true);
            convolve_separated_dxy(iy, dgau, width, dy, false);
            
            // Bild des Gradientenbetrags von dx,dy erstellen
            for (int i = 0; i < img.Height; i++)
            {
                for (int j = 0; j < img.Width; j++)
                {
                    z = norm(dx[i, j], dy[i, j]);
                    mag[i, j] = (byte)(z * MAG_SCALE);
                }
            }

            // Non-maximum suppression durchfhren
            non_max_suppress(dx, dy, mag);
            this.gradMag = mag;

            if (this.GradOnly == true)
            {
                // Kantenbild auf Original kopieren
                for (int i = 0; i < img.Height; i++)
                    for (int j = 0; j < img.Width; j++)
                        img.Data[i, j] = mag[i, j];
            }
            else
            {
                // Hysteresis thresholding anwenden
                hysteresis_thresh(img, mag);
            }
        }
        #endregion

        #region Private Methoden
        // Approximation des Betrags des Vektors (a,b) durch die Summe der Absolutbetrge.
        // -> stellt eine Approximation fr die Gradientenstrke dar...
        private float norm(float a, float b)
        {
            return (float)(Math.Abs(a) + Math.Abs(b));
        }
        
        // Berechnet den Wert der Gauss-Fkt. (in Abh. von sigma) an der Stelle x.
        private float gauss(float x, float sigma)
        {
            return (float)Math.Exp((double)((-x * x) / (2.0f * sigma * sigma)));
        }

        // Berechnet den Wert der ersten Ableitung der Gauss-Fkt. an der Stelle x.
        private float dGauss(float x, float sigma)
        {
            return -x / (sigma * sigma) * gauss(x, sigma);
        }

        // Berechnet den mittleren Wert der Gauss-Fkt. (in Abhngigkeit von sigma)
        // von 0 an der Stelle x.
        private float meanGauss(float sigma, float x)
        {
            return (gauss(x, sigma) + gauss(x + 0.5f, sigma) + gauss(x - 0.5f, sigma))
                / (3.0f * (float)Math.PI * 2.0f * sigma * sigma); 
        }

        // Faltet das Bild img mit der Maske gau in x- und y-Richtung -> Ergebnis ist ix und iy.
        private void convolve_separated(ImageGray img, float[] gau, int width, float[,] ix, float[,] iy)
        {
            float x, y;
            int i1, i2;

            for (int i = 0; i <img.Height; i++)
            {
                for (int j = 0; j < img.Width; j++)
                {
                    // Faltung mit der 1D-Gaussmaske
                    x = y = gau[0] * (float)img.Data[i, j];
                    for (int k = 1; k < width; k++)
                    {
                        i1 = (i + k) % img.Height;
                        i2 = (i - k + img.Height) % img.Height;
                        y += gau[k] * (float)img.Data[i1,j] + gau[k] * (float)img.Data[i2,j];

                        i1 = (j + k) % img.Width;
                        i2 = (j - k + img.Width) % img.Width;
                        x += gau[k] * (float)img.Data[i,i1] + gau[k] * (float)img.Data[i,i2];
                    }

                    ix[i, j] = x;
                    iy[i, j] = y;
                }
            }
        }

        // Faltet die Daten in img mit der Ableitungsmaske dgau in horizontaler Richtung, wenn horiz==true,
        // ansonsten in vertikaler Richtung. Ergebnis wird in dres geschrieben.
        private void convolve_separated_dxy(float[,] img, float[] dgau, int width, float[,] dres, bool horiz)
        {
            float res;
            int i1, i2;
            for (int i = 0; i <img.GetLength(0); i++)
            {
                for (int j = 0; j < img.GetLength(1); j++)
                {
                    res = 0.0f;
                    for (int k = 1; k < width; k++)
                    {
                        if (!horiz)
                        {
                            i1 = (i + k) % img.GetLength(0);
                            i2 = (i - k + img.GetLength(0)) % img.GetLength(0);
                            res += -dgau[k] * img[i1, j] + dgau[k] * img[i2,j];
                        }
                        else
                        {
                            i1 = (j + k) % img.GetLength(1);
                            i2 = (j - k + img.GetLength(1)) % img.GetLength(1);
                            res += -dgau[k] * img[i, i1] + dgau[k] * img[i,i2];
                        }
                    }

                    dres[i, j] = res;
                }
            }
        }

        // Fhrt Non-Max-Suppression mit den Richtungskomponenten dx,dy durch. Dabei werden mag und
        // orient auf das entstehende Betragsbild bzw. Richtungsbild gesetzt.
        private void non_max_suppress(float[,] dx, float[,] dy, byte[,] mag)
        {
            float g; // Betrag des Gradienten eines Punktes
            float g1, g2, g3, g4; // Betrag des Gradienten in versch. Richtungen
            float xc, yc; // Gradientenvektor eines Punktes
            float xx, yy; // wo wird das Nachbarpixel geschnitten?

            for (int i = 1; i < mag.GetLength(0)-1; i++)
            {
                for (int j = 1; j < mag.GetLength(1)-1; j++)
                {
                    mag[i, j] = 0;
                    // Ableitungen in x- und y-Richtung als Vektorkomponenten behandeln
                    xc = dx[i, j]; yc = dy[i, j];
                    if (Math.Abs(xc) < 0.01 && Math.Abs(yc) < 0.01)
                        continue;
                    g = norm(xc, yc);
                    
                    // folge der Gradient-Richtung
                    if (Math.Abs(yc) > Math.Abs(xc)) // steiler in y-Richtung als in x
                    {
                        xx = Math.Abs(xc / yc);
                        yy = 1.0f;

                        g2 = norm(dx[i - 1, j], dy[i - 1, j]); // Gradientenbetrag von oberem Pixel
                        g4 = norm(dx[i + 1, j], dy[i + 1, j]); // Gradientenbetrag von unterem Pixel
                        
                        if (xc * yc > 0.0) // wenn in 1. oder 3. Quadranten zeigend
                        {
                            g3 = norm(dx[i + 1, j + 1], dy[i + 1, j + 1]); // Pixel im 1. Quadranten
                            g1 = norm(dx[i - 1, j - 1], dy[i - 1, j - 1]); // Pixel im 3. Quadranten
                        }
                        else // wenn in 2. oder 4. Quadranten zeigend
                        {
                            g3 = norm(dx[i + 1, j - 1], dy[i + 1, j - 1]); // Pixel im 2. Quadranten
                            g1 = norm(dx[i - 1, j + 1], dy[i - 1, j + 1]); // Pixel im 4. Quadranten
                        }
                    }
                    else // mehr in x-Richtung zeigend als in y-Richtung
                    {
                        xx = Math.Abs(yc / xc);
                        yy = 1.0f;

                        g2 = norm(dx[i, j+1], dy[i, j+1]); // Gradientenbetrag von rechtem Pixel
                        g4 = norm(dx[i, j-1], dy[i, j-1]); // Gradientenbetrag von linkem Pixel

                        if (xc * yc > 0.0) // wenn in 1. oder 3. Quadranten zeigend
                        {
                            g3 = norm(dx[i - 1, j - 1], dy[i - 1, j - 1]); // Pixel im 3. Quadranten
                            g1 = norm(dx[i + 1, j + 1], dy[i + 1, j + 1]); // Pixel im 1. Quadranten
                        }
                        else // wenn in 2. oder 4. Quadranten zeigend
                        {
                            g1 = norm(dx[i - 1, j + 1], dy[i - 1, j + 1]); // Pixel im 4. Quadranten
                            g3 = norm(dx[i + 1, j - 1], dy[i + 1, j - 1]); // Pixel im 2. Quadranten
                        }
                    }

                    // interpolierten Wert des Gradientenbetrags berechnen
                    // wenn Gradientenbetrag g des aktuellen Pixels grer den interpolierten Nachbarbetrgen,
                    // so behalte das Pixel
                    if ((g > (xx * g1 + (yy - xx) * g2)) && (g > (xx * g3 + (yy - xx) * g4)))
                    {
                        mag[i, j] = (byte)Math.Min(255, g * MAG_SCALE);
                        this.gradOri[i, j] = (float)(Math.Atan2(yc, xc));
                    }
                    else // gehrt nicht mit zum Kantenbild
                    {
                        mag[i, j] = 0;
                        this.gradOri[i, j] = 0;
                    }
                }
            }
        }

        // Wendet Hysteresis thresholding mit den Parametern threshLow und threshHigh auf das Bild img an
        // unter Einbeziehung von mag und orient.
        private void hysteresis_thresh(ImageGray img, byte[,] mag)
        {
            // Bilddaten 0 setzen
            for (int i = 0; i < img.Height; i++)
                for (int j = 0; j < img.Width; j++)
                    img.Data[i, j] = 0;

            // Schwellwerte abschtzen
            if (this.initThreshHigh < this.initThreshLow)
            {
                estimate_thresh(mag, out this.thresholdLow, out this.thresholdHigh);
            }

            // fr jedes Pixel mit einem Betrag ber dem hohen Schwellwert:
            // verfolge Kantenpixel, die ber dem unteren Schwellwert liegen
            for (int i = 0; i < img.Height; i++)
                for (int j = 0; j < img.Width; j++)
                    if (mag[i, j] >= this.thresholdHigh)
                        trace_lowthresh(i, j, this.thresholdLow, img, mag);
        }

        private bool trace_lowthresh(int i, int j, byte threshLow, ImageGray img, byte[,] mag)
        {
            bool flag;

            if (img.Data[i, j] == 0) // sonst wurde das Pixel schon betrachtet
            {
                img.Data[i, j] = 255;
                flag = false;

                // iteriere ber Nachbarn von [i,j] und betrachte dortige Pixel wenn ntig
                for (int n = -1; n <= 1; n++)
                {
                    for (int m = -1; m <= 1; m++)
                    {
                        if (n == 0 && m == 0)
                            continue;
                        // wenn Betrag des Nachbarpixels den unteren Schwellwert bersteigt,
                        // rufe trace rekursiv auf; wenn dadurch behandelt wurde, brich Schleifen ab
                        if (indexInRange(mag, i + n, j + m) && mag[i + n, j + m] >= threshLow)
                        {
                            if (trace_lowthresh(i + n, j + m, threshLow, img, mag))
                            {
                                flag = true;
                                break;
                            }
                        }
                    }
                    if (flag == true)
                        break;
                }

                return true; // Pixel wurden behandelt
            }

            return false; // Pixel wurde schon betrachtet, wird nicht noch einmal angeschaut
        }

        // Schtzt die Werte threshLow und threshHigh anhand des bergebenen Gradientenbildes ab.
        private void estimate_thresh(byte[,] mag, out byte threshLow, out byte threshHigh)
        {
            int[] hist = new int[256]; // Histogramm
            int pxCount = mag.GetLength(0) * mag.GetLength(1);

            // Histogramm erstellen
            for (int i = 0; i < mag.GetLength(0); i++)
            {
                for (int j = 0; j < mag.GetLength(1); j++)
                    hist[mag[i, j]]++;
            }
                
            // hoher Schwellwert sollte > 80 oder 90% der Pixel sein
            int pxthresh = (int)(0.09f * pxCount);
            int k = 255;
            int count = hist[255];
            
            // berechne Schwellwert, fr den das Kriterium erfllt ist
            while (count < pxthresh)
            {
                k--;
                if (k < 0)
                {
                    k = 0;
                    break;
                }
                count += hist[k];
            }
            threshHigh = (byte)k;

            // unteren Schwellwert als Mitte zwischen oberem und niedrigstem Gradientenbetrag
            k = 0;
            while (hist[k] == 0)
                k++;
            threshLow = (byte)((threshHigh + k) / 2.0f);
        }

        // Gibt true zurck, wenn die bergebenen Indizes im Bereich des Arrays arr liegen.
        private bool indexInRange(byte[,] mag, int i, int j)
        {
            return (i >= 0 && i < mag.GetLength(0) 
                 && j >= 0 && j < mag.GetLength(1));
        }
        #endregion
    }

    /// <summary>
    /// Implementierung des optimalen Shen-Castan-Kantendetektors.
    /// </summary>
    class ShenCastanDetector : IGrayImgApplicable
    {
        #region Private Attribute
        private float smoothingFactor = 0.99f;  // Glttungsfaktor
        private byte thresholdLow = 50;          // unterer Schwellwert fr Hysteresis-thresholding
        private byte thresholdHigh = 150;        // oberer Schwellwert fr Hysteresis-thresholding
        private int windowWidth = 7;            // Fensterbreite
        private int thinningFactor = 1;         // Verdnnungsfaktor der Linien
        private int OUTLINE = 3;                // Randbereich des Bildes
        private int MAG_SCALE = 10;             // Skalierungsfaktor der Gradientenstrke
        private int threshFac = 10;              // Faktor fr Schwellwerte
        private bool gradOnly = false;          // Gradientenstrke berechnen ja/nein
        #endregion

        #region ffentliche Eigenschaften
        /// <summary>
        /// Gibt an, ob die Gradientenstrke im Bild zu speichern ist anstatt dass im Anschluss
        /// noch ein Hysteresis-thresholding durchgefhrt wird. Wenn false, dann wird das 
        /// Kantenbild binarisiert ber Hysteresis-thresholding.
        /// </summary>
        public bool GradOnly
        {
            get
            {
                return this.gradOnly;
            }
            set
            {
                this.gradOnly = value;
            }
        }
        #endregion

        #region ffentliche Methoden
        /// <summary>
        /// Erzeugt ein neues Objekt des Shen-Castan-Kantendetektors mit den bergebenen Parametern.
        /// </summary>
        /// <param name="smoothFac">Glttungsfaktor fr den ISEF-Filter</param>
        /// <param name="threshLow">Unterer Schwellwert des Hysteresis-thresholding</param>
        /// <param name="threshHigh">Oberer Schwellwert des Hysteresis-thresholding</param>
        /// <param name="winWidth">Fensterbreite fr adaptiven Gradienten - je grer, desto genauer</param>
        /// <param name="thinFac">Verdnnungsfaktor fr erkannte Kanten</param>
        public ShenCastanDetector(float smoothFac, byte threshLow, byte threshHigh, int winWidth, int thinFac)
        {
            this.smoothingFactor = Math.Max(0.0f, Math.Min(1.0f, smoothFac));
            this.thresholdLow = threshLow;
            this.thresholdHigh = threshHigh;
            this.windowWidth = winWidth;
            this.OUTLINE = winWidth / 2;
            this.thinningFactor = thinFac;
        }

        /// <summary>
        /// Wendet den Shen-Castan-Kantendetektor auf das bergebene Bild an.
        /// </summary>
        /// <param name="img">Bild, in dem Kanten detektiert werden sollen</param>
        public void Apply(ImageGray img)
        {
            // Speicher fr Teilergebnisse allokieren
            float[,] fimg = new float[img.Height, img.Width];
            byte[,]  bli  = new byte [img.Height, img.Width];

            // Eingabebild in float konvertieren
            for (int i = 0; i < img.Height; i++)
                for (int j = 0; j < img.Width; j++)
	                fimg[i,j] = (float)img.Data[i,j];

            // Eingabebild mit ISEF gltten
            float[,] isef = (new ISEFFilter(this.smoothingFactor)).CalcISEFImg(img);

            // Berechne binres Laplace-Bild
	        bli = computeBLI(isef, fimg);
            
            // Nulldurchlufe mit "false zero-crossing suppression" extrahieren und
            // adaptiven Gradienten erzeugen
            locateZeroCrossings(fimg, isef, bli);

            if (this.gradOnly == true)
            {
                // Gradientenbild speichern
                for (int i = 0; i < img.Height; i++)
                    for (int j = 0; j < img.Width; j++)
                        img.Data[i, j] = (byte)(Math.Max(0, Math.Min(255, fimg[i, j] * MAG_SCALE)));
            }
            else
            {
                // Hysteresis-Thresholding durchfhren und Ergebnis speichern
                thresholdEdges(fimg, img.Data);
                for (int i = 0; i < img.Height; i++)
                    for (int j = 0; j < img.Width; j++)
                        img.Data[i, j] = (byte)((img.Data[i, j] > 0) ? 255 : 0);
            }
        }
        #endregion

        #region Private Methoden
        // berechne binres Laplace-Bild (BLI)
        private byte[,] computeBLI (float[,] isef, float[,] orig)
        {
	        int height = orig.GetLength(0);
            int width  = orig.GetLength(1);
            byte[,] bli = new byte[height,width];
           
	        // das BLI wird berechnet, indem die Differenz von ISEF-Bild und Original-Bild
            // berechnet wird. Dies ergibt das band-begrenzte Laplace-Bild. Dieses wird dann
            // binarisiert, indem alle positiven Werte auf 1 und alle negativen Werte auf 0 
            // gesetzt werden.
            for (int row = 0; row < height; row++)
                for (int col = 0; col < width; col++)
                {
                    if (row < OUTLINE || row >= height - OUTLINE ||
                        col < OUTLINE || col >= width - OUTLINE) continue;
                    bli[row, col] = (byte)(((isef[row, col] - orig[row, col]) > 0.0f) ? 1 : 0);
                }
	        return bli;
        }

        // Nulldurchlufe im BLI erkennen
        private void locateZeroCrossings (float[,] orig, float[,] smoothed, byte[,] bli)
        {
            int height = orig.GetLength(0);
            int width  = orig.GetLength(1);

	        for (int row = 0; row < height; row++)
            {
                for (int col = 0; col < width; col++)
	            {

                    // Pixel am Bildrand ignorieren
	                if (row<OUTLINE || row >= height-OUTLINE ||
		               col<OUTLINE || col >= width-OUTLINE)
	                {
	                    orig[row,col] = 0.0f;
	                }

                   // prfe ob das Pixel ein Nulldurchlauf des Laplace-Bildes ist
	               else if (isCandidateEdge (bli, smoothed, row, col))
	               {
                       // nun wende adaptiven Gradienten an
	                   float grad = computeAdaptiveGradient (bli, smoothed, row, col);
                       orig[row, col] = grad;
	               }
	               else
                       orig[row,col] = 0.0f;		    
	            }
	        }
        }

        // Hysteresis-Thresholding durchfhren
        private void thresholdEdges (float[,] grad, byte[,] edges)
        {
	        int height = grad.GetLength(0);
            int width = grad.GetLength(1);

            byte[,] igrad = new byte[height, width];
            for (int i = 0; i < height; i++)
            {
                for (int j = 0; j < width; j++)
                {
                    igrad[i, j] = (byte)grad[i, j];
                    edges[i, j] = 0;
                }
            }

            for (int i = 0; i < height; i++)
                for (int j = 0; j < width; j++)
	            {
                    if (i<OUTLINE || i >= height-OUTLINE ||
                        j<OUTLINE || j >= width-OUTLINE) continue;

                    // Kontur nur verfolgen, wenn sie ber oberem Schwellwert liegt
                    if ((grad[i, j]) * this.threshFac > (float)this.thresholdHigh)
                        // markiere alle miteinander verbundenen Punkte ber unterem Schwellwert
                        markConnected(i, j, 0, this.thresholdLow, grad, edges);
	            }

            // lsche alle Punkte deren Wert 255 betrgt
            for (int i = 0; i < height; i++)
                for (int j = 0; j < width; j++)
	                if (edges[i,j] == 255) 
                        edges[i,j] = 0;
        }

        // Markieren aller miteinander verbundenen Punkte, die ber dem unteren Schwellwert liegen.
        // Gibt true zurck, wenn etwas markiert wurde.
        private int markConnected (int i, int j, int level, float threshLow, float[,] grad, byte[,] edges)
        {
	        int notChainEnd;

            int height = grad.GetLength(0);
            int width = grad.GetLength(1);
            
            // stoppen wenn Bereichsgrenzen verlassen werden
            if (i >= height || i < 0 || j >= width || j < 0)
                return 0;
           
            // stoppen wenn dieses Pixel bereits betrachtet wurde
	        if (edges[i,j] != 0)
                return 0;	
           
            // wenn Gradient 0 ist, brauchen wir auch nicht weiter machen
            if (grad[i,j] == 0.0f)
                return 0;
           
            // wenn Gradient ber unterem Schwellwert, dann weiter betrachten
            edges[i, j] = (byte)(((grad[i, j]) * this.threshFac > (float)threshLow) ? 1 : 255);

            notChainEnd = 0;

            // jedes Pixel in Achternachbarschaft betrachten
            notChainEnd |= markConnected(i, j + 1, level + 1, threshLow, grad, edges);
            notChainEnd |= markConnected(i, j - 1, level + 1, threshLow, grad, edges);
            notChainEnd |= markConnected(i + 1, j + 1, level + 1, threshLow, grad, edges);
            notChainEnd |= markConnected(i + 1, j, level + 1, threshLow, grad, edges);
            notChainEnd |= markConnected(i + 1, j - 1, level + 1, threshLow, grad, edges);
            notChainEnd |= markConnected(i - 1, j - 1, level + 1, threshLow, grad, edges);
            notChainEnd |= markConnected(i - 1, j, level + 1, threshLow, grad, edges);
            notChainEnd |= markConnected(i - 1, j + 1, level + 1, threshLow, grad, edges);

            if (notChainEnd > 0 && level > 0)
            {
                // Kontur-Verdnnung
                if (this.thinningFactor > 0.0f)
                    if (level % thinningFactor != 0)
                        edges[i, j] = 255; // lsche diesen Punkt
            }
            
	        return 1;
        }

        // Finde Nulldurchlufe im BLI, orig ist das geglttete Bild.
        private bool isCandidateEdge (byte[,] bli, float[,] orig, int row, int col)
        {
            // Teste auf Nulldurchlauf im Laplace-Bild, dann berprfe Gltigkeit
            // des Nulldurchlaufs mit Gradienten-Vorzeichen.
            if (bli[row, col] == 1 && bli[row + 1, col] == 0) // positiver Nulldurchlauf
                return (orig[row + 1, col] - orig[row - 1, col] > 0.0f) ? true : false;
            else if (bli[row, col] == 1 && bli[row, col + 1] == 0) // positiver Nulldurchlauf
                return (orig[row, col + 1] - orig[row, col - 1] > 0.0f) ? true : false;
            else if (bli[row, col] == 1 && bli[row - 1, col] == 0) // negativer Nulldurchlauf
                return (orig[row + 1, col] - orig[row - 1, col] < 0.0f) ? true : false;
            else if (bli[row, col] == 1 && bli[row, col - 1] == 0) // negativer Nulldurchlauf
                return (orig[row, col + 1] - orig[row, col - 1] < 0.0f) ? true : false;
            else // kein Nulldurchlauf
                return false;
        }

        // Wert des adaptiven Gradienten an der Stelle (row,col) berechnen
        private float computeAdaptiveGradient (byte[,] bli, float[,] orig, int row, int col)
        {
	        float sumOn, sumOff;
	        float avgOn, avgOff;
	        int numOn, numOff;
           
	        sumOn = sumOff = 0.0f;
	        numOn = numOff = 0;
            
	        // betrachte jedes Pixel unter dem Fenster
            for (int i = -this.windowWidth / 2; i <= this.windowWidth / 2; i++)
            {
                for (int j = -this.windowWidth / 2; j <= this.windowWidth / 2; j++)
                {
                    // erste Region hat positive Werte im BLI
                    if (bli[row + i, col + j] > 0)
                    {
                        sumOn += orig[row + i, col + j];
                        numOn++;
                    }
                    // zweite Region hat negative Werte im BLI
                    else
                    {
                        sumOff += orig[row + i, col + j];
                        numOff++;
                    }
                }
            }
           
	        avgOff = (sumOff>0.0f) ? sumOff/(float)numOff : 0.0f;
            avgOn = (sumOn>0.0f) ? sumOn/(float)numOn : 0.0f;
           
            // Gradientenbetrag ergibt sich aus der Differenz der beiden Regionen
	        return (avgOff - avgOn);
        }
        #endregion
    }

    /// <summary>
    /// Implementierung der optimalen Marr-Hildreth-Kantendetektors (Laplacian-of-Gaussian, LoG).
    /// In der vorliegenden Version recht lahm ^^.
    /// </summary>
    class MarrHildrethDetector : IGrayImgApplicable
    {
        #region Private Attribute
        private float sigma = 1;        // Standardabweichung zur Berechnung der Gauss-Funktion
        #endregion

        #region ffentliche Methoden
        /// <summary>
        /// Erzeugt einen neuen Marr-Hildreth-Kantendetektor mit der bergebenen Standardabweichung
        /// fr die Gauss-Funktion.
        /// </summary>
        /// <param name="sigma">Standardabweichung der Gauss-Funktion, minimal 1.0</param>
        public MarrHildrethDetector(float sigma)
        {
            this.sigma = (float)Math.Max(1.0, sigma);
        }

        /// <summary>
        /// Wendet den Marr-Hildreth-Kantendetektor auf das bergebene Bild an.
        /// </summary>
        /// <param name="img">Bild, in dem Kanten detektiert werden sollen</param>
        public void Apply(ImageGray img)
        {
            ImageGray img2 = new ImageGray(img);
            
            // Operator mit sigma-Werten zu beiden Seiten des eigentlichen Wertes anwenden
            marr((float)(sigma - 0.8), img);
            marr((float)(sigma + 0.8), img2);

            // beide Bilder auswerten
            for (int y = 0; y < img.Height; y++)
                for (int x = 0; x < img.Width; x++)
                    // nur auf 255 setzen, wenn beide LoG-Bilder > 0
                    if (img.Data[y, x] > 0 && img2.Data[y, x] > 0)
                        img.Data[y, x] = (byte)255;
                    else
                        img.Data[y, x] = (byte)0;
        }
        #endregion

        #region Private Methoden
        // Gibt die Lnge des bergebenen Vektors zurck.
        private float norm(float x, float y)
        {
            return (float)Math.Sqrt((double)(x * x + y * y));
        }

        // Gibt den Abstand der beiden bergebenen Punkte zurck.
        private float distance(float x1, float y1, float x2, float y2)
        {
            return norm((x1 - x2), (y1 - y2));
        }

        // Fhrt die LoG-Filterung des bergebenen Bildes durch.
        private void marr(float sigma, ImageGray img)
        {
            int width, n;
            float[,] smx = new float[img.Height, img.Width];
            float[,] lgau;

            // erzeuge Gauss-Filtermaske sowie deren Ableitung
            width = (int)(3.35 * sigma + 0.33);
            n = 2 * width + 1;
            lgau = new float[n,n];
            for (int i = 0; i < n; i++)
                for (int j = 0; j < n; j++)
                    lgau[i,j] = lapOfGauss(distance((float)i, (float)j,
                         (float)width, (float)width), sigma);

            // falte Ausgangsbild mit Gauss in X und Y Richtung
            convolution(img, lgau, smx);

            // ermittle Nulldurchlufe
            zero_cross(smx, img);
        }

        // Berechnet den Wert der Gauss-Funktion an der Stelle x mit der bergebenen Standardabweichung sigma.
        private float gauss(float x, float sigma)
        {
            return (float)Math.Exp((double)((-x * x) / (2.0 * sigma * sigma)));
        }

        // Berechnet den mittleren Wert der Gauss-Funktion an der Stelle x.
        private float meanGauss(float x, float sigma)
        {
            double z = (gauss(x, sigma) + gauss((float)(x + 0.5), sigma) + gauss((float)(x - 0.5), sigma)) / 3.0;
            return (float)(z / (Math.PI * 2.0 * sigma * sigma));
        }

        // Berechnet den Laplace-Wert der Gauss-Funktion an der gegebenen x-Koordinate.
        private float lapOfGauss(float x, float sigma)
        {
            float x1 = gauss(x, sigma);
            return (x * x - 2 * sigma * sigma) / (sigma * sigma * sigma * sigma) * x1;
        }

        // Faltet das bergebene Bild mit der angegebenen Maske, Ergebnis wird in res gespeichert.
        private void convolution(ImageGray img, float[,] mask, float[,] res)
        {
            int n, m, k, kk;
            float x;

            // lineare Faltung
            k = mask.GetLength(0) / 2; kk = mask.GetLength(1) / 2;
            for (int i = 0; i < img.Height; i++)
                for (int j = 0; j < img.Width; j++)
                {
                    x = 0.0f;
                    for (int ii = 0; ii < mask.GetLength(0); ii++)
                    {
                        n = i - k + ii;
                        if (n < 0 || n >= img.Height) continue;
                        for (int jj = 0; jj < mask.GetLength(1); jj++)
                        {
                            m = j - kk + jj;
                            if (m < 0 || m >= img.Width) continue;
                            x += mask[ii,jj] * (float)(img.Data[n,m]);
                        }
                    }
                    res[i,j] = x;
                }
        }

        // Erkennung von Nulldurchlufen im Laplace-Bild, Ergebnis-Speicherung im bergebenen Grauwertbild.
        private void zero_cross(float[,] lapImg, ImageGray img)
        {
            for (int i = 1; i < img.Height-1; i++)
                for (int j = 1; j < img.Width-1; j++)
                {
                    img.Data[i,j] = 0;
                    // prfe auf Nulldurchlauf im Laplace-Bild
                    if (lapImg[i - 1, j] * lapImg[i + 1, j] < 0
                     || lapImg[i, j - 1] * lapImg[i, j + 1] < 0
                     || lapImg[i + 1, j - 1] * lapImg[i - 1, j + 1] < 0
                     || lapImg[i - 1, j - 1] * lapImg[i + 1, j + 1] < 0)
                    {
                        img.Data[i, j] = (byte)255;
                        continue;
                    }
                }
        }
        #endregion
    }

    /// <summary>
    /// In dieser Klasse werden statische get-Properties fr die optimalen Kantendetektoren
    /// mit voreingestellten Parametern gespeichert.
    /// </summary>
    sealed class OptimalEdgeDetectors
    {
        #region ffentliche Eigenschaften
        /// <summary>
        /// get-Property fr ein Objekt des Marr-Hildreth-Kantendetektors mit
        /// einem Wert fr sigma von 2.
        /// </summary>
        public static MarrHildrethDetector MarrHildreth
        {
            get
            {
                return new MarrHildrethDetector(2.0f);
            }
        }
        
        /// <summary>
        /// get-Property fr ein Objekt des Canny-Kantendetektors mit einem sigma-Wert von 2
        /// und automatischer Schwellwertberechnung. Das Kantenbild wird binarisiert ausgegeben.
        /// </summary>
        public static CannyDetector Canny
        {
            get
            {
                return new CannyDetector(2.0f, 70, 60, false);
            }
        }

        /// <summary>
        /// get-Property fr ein Objekt des Shen-Castan-Kantendetektors mit einem sigma-Wert
        /// von 0.6, einem unteren Schwellwert von 20 und einem oberen von 10, einer
        /// Fensterbreite von 5 und keiner Linienverdnnung.
        /// </summary>
        public static ShenCastanDetector ShenCastan
        {
            get
            {
                return new ShenCastanDetector(0.6f, 20, 70, 5, 1);
            }
        }
        #endregion
    }
}
